#
# scalar.py
#
# Copyright (c) 2016 Junpei Kawamoto
#
# This file is part of rgmining-review.
#
# rgmining-review is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# rgmining-review is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
"""Implementations of scalar review and summary classes.
"""
from __future__ import absolute_import
import numbers
import numpy as np

from review.base import Review
from review.base import Summary


class AverageReview(Review):
    """Scalar review.

    Args:
      v: a float value representing review score.

    The review score is a scalar value.
    """
    __slots__ = ("_v")

    def __init__(self, v):
        """Construct average review class.

        Args:
          v: a float value representing review score.
        """
        if not isinstance(v, numbers.Number):
            raise TypeError("v ({0}) is not a scalar value.".format(type(v)))
        self._v = float(v)

    @property
    def score(self):
        """A float value representing score of this review. """
        return self._v

    def __eq__(self, other):
        if not isinstance(other, AverageReview):
            return False
        return self.score == other.score

    def __add__(self, other):
        if not isinstance(other, AverageReview):
            raise TypeError(
                "other is {0}, not AverageReview".format(type(other)))
        return AverageReview(self.score + other.score)

    def __rmul__(self, other):
        if not isinstance(other, numbers.Number):
            raise TypeError(
                "other is {0}, not numbers.Number".format(type(other)))
        return AverageReview(other * self.score)

    def __str__(self):
        return str(self.score)

    def __hash__(self):
        return hash(self.score)


class AverageSummary(Summary):
    """Scalar summary.

    The summary is an average of given reviews.
    """
    __slots__ = ("_v")  # _v : an instance of AgerageReview

    def __init__(self, scores):
        if hasattr(scores, "__iter__"):
            v = np.mean(list(scores))
            if isinstance(v, AverageReview):
                self._v = v
            else:
                self._v = AverageReview(v)
        elif isinstance(scores, AverageReview):
            self._v = scores
        else:
            self._v = AverageReview(scores)

    def difference(self, r):
        """Difference between this summary and a given review.

        Args:
          r: a review.

        Returns:
          a non-negative float value or 0 representing the difference between
          this summary and the given value.
        """
        if not isinstance(r, AverageReview):
            raise TypeError("r is {0}, not AverageReview".format(type(r)))
        return abs(self._v.score - r.score)

    @property
    def score(self):
        """Float value representing this summary.
        """
        return self._v.score

    @property
    def v(self):
        """Summary score.
        """
        return self._v.score

    def __str__(self):
        return str(self._v)

    @classmethod
    def review_class(cls):
        """A review class associated with this summary. """
        return AverageReview



#
# histogram.py
#
# Copyright (c) 2016 Junpei Kawamoto
#
# This file is part of rgmining-review.
#
# rgmining-review is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# rgmining-review is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
"""Implementations of histogram review and summary classes.
"""
from __future__ import absolute_import
from collections import defaultdict
import numbers

import numpy as np

from review.base import Review
from review.base import Summary


class HistoReview(Review):
    """Vector Review.
    """
    __slots__ = ("_v")

    def __init__(self, v, quantizer=round):
        self._v = {}
        if isinstance(v, dict):
            for key, value in v.items():
                self._v[quantizer(key)] = float(value)
        else:
            self._v[quantizer(v)] = 1.

    @property
    def score(self):
        """A float value representing score of this review. """
        res = 0.
        for k in self:
            res += k * self[k]
        return res

    @property
    def vector(self):
        """ Raw vector.
        """
        return self._v

    def norm(self):
        """ 1-Norm of this vector.
        """
        return sum(self._v.values(), 0.)

    def inner_product(self, other):
        """ Inner product of two vectors.

        Args:
          other: a HistogramReview instance.

        Returns:
          the inner product between this and the other.
        """
        if not isinstance(other, HistoReview):
            raise TypeError(
                "other must be an HistoReview: {0}".format(type(other)))
        res = 0.
        for k in set(self.vector.keys()) & set(other.vector.keys()):
            res += self.vector[k] * other.vector[k]
        return res

    def __eq__(self, other):
        if not isinstance(other, HistoReview):
            return False
        if self.vector.keys() != other.vector.keys():
            return False
        for k in self:
            if self[k] != other[k]:
                return False
        return True

    def __add__(self, other):
        if not isinstance(other, HistoReview):
            raise TypeError(
                "other is {0}, not HistoReview".format(type(other)))
        res = defaultdict(float, self._v)
        for k in other:
            res[k] += other[k]
        return HistoReview(res)

    def __rmul__(self, other):
        if not isinstance(other, numbers.Number):
            raise TypeError("other is {0}, not a number".format(type(other)))
        res = self._v.copy()
        for k in res:
            res[k] *= other
        return HistoReview(res)

    def __getitem__(self, key):
        return self._v[key]

    def __iter__(self):
        return self._v.__iter__()

    def __contains__(self, v):
        return v in self._v

    def __str__(self):
        return ", ".join(["{0}:{1}".format(i, self[i]) for i in self])


class HistoSummary(Summary):
    """ Vector summary.
    """
    __slots__ = ("_histo")  # _histo: an instance of HistoReview

    def __init__(self, reviews):
        if hasattr(reviews, "__iter__"):
            reviews = list(reviews)
            if not isinstance(reviews[0], HistoReview):
                reviews = [HistoReview(r) for r in reviews]
            self._histo = np.mean(reviews)
        elif isinstance(reviews, HistoReview):
            self._histo = reviews
        else:
            self._histo = HistoReview(reviews)

    def difference(self, r):
        """Compute a difference between this summary and a given review score.

        Args:
          An instance of Review.

        Returns:
          The difference between of the summary and the given review.
        """
        if not isinstance(r, HistoReview):
            raise TypeError("r must be an HistoReview: {0}".format(type(r)))
        return abs(1 - self._histo.inner_product(r))

    @property
    def score(self):
        """Return a float value representing this summary.
        """
        return self._histo.score

    def __str__(self):
        return ", ".join(["{0}:{1}".format(i, self._histo[i]) for i in self._histo])

    @classmethod
    def review_class(cls):
        """A review class associated with this summary. """
        return HistoReview



#
# base.py
#
# Copyright (c) 2016 Junpei Kawamoto
#
# This file is part of rgmining-review.
#
# rgmining-review is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# rgmining-review is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
"""Defines abstract Review and Summary classes.
"""
import math
import numbers


class _ImmutableAdditiveGroup(object):
    """Immutable additive group.

    Subclass must implement __add__, __neg__, and __eq__, then this class
    complements __sub__ and __ne__.
    """
    __slots__ = ()

    def __add__(self, _):
        raise NotImplementedError("Subclasses must implement __add__.")

    def __sub__(self, other):
        return self + (-other)

    def __neg__(self):
        raise NotImplementedError("Subclasses must implement __neg__.")

    def __eq__(self, _):
        raise NotImplementedError("Subclasses must implement __eq__.")

    def __ne__(self, other):
        return not self == other


class _MultipliableImmutableAdditiveGroup(_ImmutableAdditiveGroup):
    """Multipliable immutable additive group.

    This is a subclass of _ImmutableAdditiveGroup.
    Subclass must implement __add__, __rmul__, and __eq__, then this class
    complements __div__ and __neg__.
    """
    __slots__ = ()

    def __rmul__(self, _):
        # value must be a number.
        raise NotImplementedError("Subclasses must implement __rmul__")

    def __div__(self, value):
        return self.__truediv__(value)

    def __truediv__(self, value):
        if not isinstance(value, numbers.Number):
            raise TypeError("value must be an instance of numbers.Number")
        return (1. / value) * self

    def __floordiv__(self, value):
        if not isinstance(value, numbers.Number):
            raise TypeError("value must be an instance of numbers.Number")
        return math.floor((1. / value) * self)

    def __neg__(self):
        return -1 * self


class Review(_MultipliableImmutableAdditiveGroup):
    """Abstruct class of Review.

    Review is defined on a multipliable immutable additive group.
    Subclass must implement the following methods;

    - _eq_:
    - _add_:
    - _rmul_:

    Review also needs to implement a property `score` to return the review
    score itselt. The returned score must be a float number.
    """
    __slots__ = ()

    def __mul__(self, other):
        return self.__rmul__(other)

    @property
    def score(self):
        """A float value representing score of this review. """
        raise NotImplementedError


class Summary(object):
    """ Abstract class of summary of reviews.

    Summary only needs to define `differenct` and `norm` methods.
    `difference` computes difference between a summary and a review.
    `norm` maps a summary to a float value.

    Each summary type might be related to a review type.
    Summary must implements a class method `review_class` to return the
    associated review class.
    """
    __slots__ = ()

    def difference(self, r):
        """Compute a difference between this summary and a given review score.

        Args:
          An instance of Review.

        Returns:
          The difference between of the summary and the given review.
        """
        raise NotImplementedError

    @property
    def score(self):
        """Return a float value representing this summary.
        """
        raise NotImplementedError

    @classmethod
    def review_class(cls):
        """A review class associated with this summary. """
        raise NotImplementedError



#
# setup.py
#
# Copyright (c) 2016 Junpei Kawamoto
#
# This file is part of rgmining-review.
#
# rgmining-review is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# rgmining-review is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
"""Package information of review package for review graph mining project.
"""
from setuptools import setup, find_packages


def _load_requires_from_file(filepath):
    """Read a package list from a given file path.

    Args:
      filepath: file path of the package list.

    Returns:
      a list of package names.
    """
    with open(filepath) as fp:
        return [pkg_name.strip() for pkg_name in fp.readlines()]


setup(
    name='rgmining-review',
    version='0.9.0',
    author="Junpei Kawamoto",
    author_email="kawamoto.junpei@gmail.com",
    description="Review data structure for Review graph mining project",
    url="https://github.com/rgmining/common",
    packages=find_packages(exclude=["tests"]),
    install_requires=_load_requires_from_file("requirements.txt"),
    test_suite='tests.suite',
    license="GPLv3",
    classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Natural Language :: English",
            "Programming Language :: Python",
            "Topic :: Software Development :: Libraries",
            "Topic :: Scientific/Engineering :: Information Analysis"
        ]
)



#
# __init__.py
#
# Copyright (c) 2016 Junpei Kawamoto
#
# This file is part of rgmining-review.
#
# rgmining-review is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# rgmining-review is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
"""Review data model consisting of review and summary.

This module defines abstract :class:`review.base.Review` and
:class:`review.base.Summary`, and those implementations.
:class:`review.base.Review` represents review itself and
:class:`review.base.Summary` is a summarized reviews given to a same object.

This module assumes each review is representable in one scalar value,
and those review values are in a multipliable additive group.
With the assumption, this module implements two types of summarization methods.
One of them is average of reviews and the other one is making a vector,
in other words histogram, of reviews.

This module has the following aliases:

- :class:`AverageReview <review.scalar.AverageReview>`
- :class:`AverageSummary <review.scalar.AverageSummary>`
- :class:`HistoReview <review.histogram.HistoReview>`
- :class:`HistoSummary <review.histogram.HistoSummary>`

"""
from __future__ import absolute_import
from review.scalar import AverageReview
from review.scalar import AverageSummary
from review.histogram import HistoReview
from review.histogram import HistoSummary
